/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Pusher 0.6.1 -- text mode Sokoban game.
; conv.c -- maze converter.
; Copyright (c) 2009  Oleg O. Chukaev  <oleg.chukaev@mail.ru>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
; 02110-1301 USA
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/


#include <stdio.h>
#include <stdlib.h>


FILE *f;
int db[500], de[500];
int db_i, de_i, my_x, my_y, rows, cols;


int conv (char c) {
	switch (c) {
	case '$': return 0;
	case '.': return 0;
	case 'B': return 1;
	case '#': return 2;
	case '-': return 3;
	case 'x': return 4;
	case 'b': return 5;
	}
	return 0;
}


int get_maze (void) {

	int i, j, cnt, z, ch;
	char str[1000];

	i = fscanf (f, "%d, %d", &cols, &rows);
	if (i != 2)
		return 1;

	db_i = 0;
	cnt = 1;
	z = 0;
	for (i = 0; i < rows; i++) {
		fscanf (f, "%s", str);
		for (j = 0; j < cols; j++) {
			ch = conv (str[j]);
			if (str[j] == '$') {
				my_x = j; my_y = i;
			}
			z += ch * cnt;
			cnt *= 6;
			if (cnt == 6 * 6 * 6) {
				db[db_i] = z;
				db_i++;
				cnt = 1;
				z = 0;
			} else if (i == rows - 1 && j == cols - 1) {
				db[db_i] = z;
				db_i++;
			}
		}
	}
	return 0;
}


void encode (void) {

	int ch, cnt, i;

	ch = db[0];
	cnt = 1;
	de_i = 0;
	for (i = 1; i < db_i; i++) {
		if (db[i] == ch) {
			cnt++;
		} else {
			if (cnt == 1) {
				de[de_i++] = ch;
				ch = db[i];
			} else {
				de[de_i++] = 0xe0 + cnt;
				de[de_i++] = ch;
				ch = db[i];
				cnt = 1;
			}
		}
	}
	de[de_i++] = ch;
}


void save (int num) {

	int i, j;

	printf ("maze%02d:\n", num);
	printf ("\t\tdb\t%d, %d\n", cols, rows);
	printf ("\t\tdb\t%d, %d\n", my_x, my_y);
	for (i = 0; i < de_i - 8; i+=8) {
		printf ("\t\tdb\t0%02xh, 0%02xh, 0%02xh, 0%02xh, 0%02xh, "
			"0%02xh, 0%02xh, 0%02xh\n", de[i], de[i+1], de[i+2],
			de[i+3], de[i+4], de[i+5], de[i+6], de[i+7]);
	}
	if (i < de_i) {
		printf ("\t\tdb\t");
		for (j = 0; j < de_i - i; j++) {
			printf ("0%02xh%s", de[i+j], (j == de_i - i - 1) ?
				"\n" : ", ");
		}
	}
	printf ("\n");
}


int main (void) {

	int i, m;

	f = fopen ("maze.dat", "rt");
	if (f == NULL) {
		perror ("Can't open 'maze.txt'. . .\n");
		return (1);
	}

	printf ("; Generated automagically, do not edit...\n\n");
	m = 1;
	while (1) {
		if (get_maze ())
			break;
		encode ();
		save (m++);
	}
	printf ("\nmazes:\n");
	for (i = 1; i < m; i++) {
		printf ("\t\tdw\tmaze%02d\n", i);
	}
	printf ("end_of_mazes:\n\n");
	printf ("; E0F\n");

	fclose (f);
	return 0;
}


/* E0F */
