;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Pusher 0.6.1 -- text mode Sokoban game.
; Copyright (c) 2009  Oleg O. Chukaev  <oleg.chukaev@mail.ru>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
; 02110-1301 USA
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Constants
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
%define		RELEASE	"0.6.1"		;Version number
CR		equ	0dh		;Carriage Return
LF		equ	0ah		;Linefeed

LAST_MAZE	equ	(end_of_mazes - mazes) / 2 - 1

SPACE		equ	00h		;
BOX0		equ	01h		;
WALL		equ	02h		;
EMPTY		equ	03h		;
TARGET		equ	04h		;SPACE+W_MASK
BOX1		equ	05h		;BOX0+W_MASK
FRAME		equ	06h		;Frame
MAN		equ	07h		;\ Must be consecutive
MAN2		equ	08h		;/

W_MASK		equ	04h

VIDEO_SEG	equ	0b800h		;Segment of video buffer
KEY_LEFT	equ	4bh		;------------
KEY_RIGHT	equ	4dh		;
KEY_UP		equ	48h		;
KEY_DOWN	equ	50h		;Scancodes
KEY_ESC		equ	01h		;
KEY_SPACE	equ	39h		;
KEY_BKSP	equ	0eh		;------------

COL_FRAME	equ	70h		;Color of frame
COL_HELP	equ	13h		;Color of help line
COL_STATUS	equ	71h		;Color of status msgs
COL_BG1		equ	02h		;1st background color
COL_BG2		equ	0ch		;2nd background color
COL_EMPTY	equ	13h		;Color of empty space
COL_TARGET	equ	2fh		;Color of targets
COL_MAN		equ	0eh		;Color of the man
COL_BOX0	equ	4eh		;Color of box at empty space
COL_BOX1	equ	5fh		;Color of box at target
COL_WALL	equ	70h		;Color of wall

MAX_FIELD_W	equ	32		;Max width of the game field
MAX_FIELD_H	equ	20		;Max height of the game field
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
		section	.text align=1
		org	100h
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Code
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

;=============================================================================
; main
;-----------------------------------------------------------------------------
; Features of initialization code:
; a) used segment 0b800h, i.e. monochrome adapter not supported;
; b) used 16 scanlines per char: 80x50 will be changed to 80x25,
;    80x30 supported, but last 5 lines not used.
; c) used 80 chars per line: 40x25 will be changed to 80x25
;
; In:	---
; Out:	---
; Modf:	AX, BX, CX, DX, DI, SI
; Call:	charxchg, init_map, print_frame, print_maze, game
; Use:	copyright, curshape, rgb, randnum, save_buffer, moves, pushes
;
main:
		mov	dx,copyright	;Copyright message
		mov	ah,09h
		int	21h
;-----------------------------------------------------------------------------
; Save cursor shape, palette, screen contents
;-----------------------------------------------------------------------------
		xor	ax,ax
		mov	es,ax
		cmp	word [es:485h],16
		jne	@@set_mode_3	;If not 16 scanlines per char...
		cmp	word [es:44ah],80
		jne	@@set_mode_3	;...or not 80 columns then set mode 03
		cmp	byte [es:484h],24
		jae	@@good_screen	;If >= 25 lines then mode not changed
@@set_mode_3:
		mov	al,03h
		int	10h		;AH == 0 after xor ax,ax
@@good_screen:
		push	cs
		pop	es
		mov	ah,03h
		mov	bx,0001h	;Video page for fn 03h / blue color
		int	10h		;for fn 1015h
		mov	[curshape],cx	;Save cursor shape
		mov	ax,1015h	;Save RGB
		int	10h
		mov	[rgb+0],dh	;R
		mov	[rgb+1],cx	;G, B
;-----------------------------------------------------------------------------
		mov	ah,0		;Randomize
		int	1ah
		mov	[randnum],cx	;?Maybe DX?
		mov	[randnum+2],dx	;?Maybe CX?
;-----------------------------------------------------------------------------
		mov	di,save_buffer	;Buffer for screen
		mov	ax,VIDEO_SEG
		push	ax
		mov	ds,ax
		xor	si,si		;DS:SI -> beginnig of videobuffer
		mov	cx,25*80	;25 lines, 80 columns
	rep	movsw
		push	cs
		pop	ds
;-----------------------------------------------------------------------------
; Set cursor shape, palette, font
;-----------------------------------------------------------------------------
		mov	ah,01h
		mov	cx,200ah	;Invisible cursor, blue -> dark blue
		int	10h		;^---- CH, CL -----^
		mov	ax,1010h
		mov	ch,0		;G
		mov	dh,0		;R
		int	10h
		call	charxchg	;Set new font
		pop	es		;ES == VIDEO_SEG
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
@@next_game:
		mov	word [moves],0
		mov	word [pushes],0
		call	init_map
		call	print_frame
		call	print_maze
		call	game		;Game
		jnc	@@next_game	;ESC not pressed -- paly again
;-----------------------------------------------------------------------------
; Restore screen, cursor, font, palette, and exit
;-----------------------------------------------------------------------------
@@exit:
		mov	ah,01h
		mov	cx,[curshape]	;Restore:
		int	10h		;cursor shape,
		mov	dh,[rgb]	;R
		mov	cx,[rgb+1]	;G, B
		mov	ax,1010h
		mov	bx,1
		int	10h		;palette,
		xor	di,di
		mov	si,save_buffer
		mov	cx,25*80
	rep	movsw			;screen,
		call	charxchg	;and fonts
		ret			;Return to DOS
;=============================================================================
; game
;-----------------------------------------------------------------------------
; Main game procedure.
; In:	---
; Out:	CF == 1 if exit from program required
; Modf:	AX, ~BX, CX, DX, ~DI, SI
; Call:	showstatus, fill_bg, changefield, move, print_maze, undo
; Use:	undo_enabled, starttime, boxes, in_place, directions
;
game:
		mov	byte [undo_enabled],0
@@waitkey:
		mov	ah,1
		int	16h
		jz	@@waitkey

		mov	ah,0
		int	1ah
		mov	[starttime],dx
		mov	[starttime+2],cx
@@readkey:
		call	showstatus	;Show status of the game
		mov	al,[boxes]
		cmp	al,[in_place]
		jne	@@checkkey
		push	ds
		xor	ax,ax
		mov	ds,ax
		mov	ax,[46ch]
		mov	dx,[46eh]
		add	ax,10		;~0.5 sec.
		adc	dx,0
@@wait:
		cmp	ax,[46ch]
		ja	@@wait
		cmp	dx,[46eh]
		ja	@@wait
		pop	ds

		call	fill_bg

		mov	al,'+'
		call	changefield
		jmp	@@quit		;Quit, carry flag clear
@@checkkey:
		mov	ah,1
		int	16h
		jz	@@readkey
		mov	ah,0
		int	16h
		mov	cl,al
		mov	si,directions	;Shift table
@@next_scan:
		lodsb
		or	al,al		;End of table?
		jz	@@not_found	;Yes, go check another keys
		cmp	al,ah
		je	@@found
		cmpsw			;Skip additions to DL and DH
		jmp	@@next_scan	;\__OPTIMIZE: instead of INC SI; INC SI
@@found:
		lodsw			;Additions to cursor coordinates
		call	move
		call	print_maze
		jmp	@@readkey
@@not_found:
		mov	al,cl
		cmp	al,'+'
		je	@@changefield
		cmp	al,'='
		je	@@changefield
		cmp	al,'-'
		jne	@@check_sp
@@changefield:
		call	changefield
		jc	@@readkey
		jmp	@@quit

@@check_sp:
		cmp	ah,KEY_SPACE
		je	@@quit
		cmp	al,'r'
		je	@@quit
		cmp	al,'R'
		je	@@quit
@@check_bksp:
		cmp	ah,KEY_BKSP
		je	@@undo
		cmp	al,'u'
		je	@@undo
		cmp	al,'U'
		jne	@@esc
@@undo:
		call	undo
		call	print_maze
		jmp	@@readkey
@@esc:
		cmp	ah,KEY_ESC
		je	@@quit_wc	;Set carry flag and quit
		cmp	al,'q'
		je	@@quit_wc	;Set carry flag and quit
		cmp	al,'Q'
		je	@@quit_wc	;Set carry flag and quit
		jmp	@@readkey
@@quit:
		clc
		ret
@@quit_wc:
		stc
		ret
;=============================================================================
; save_map
;-----------------------------------------------------------------------------
; Saves game map.
; In:	---
; Out:	---
; Modf:	CX, SI, DI
; Call:	---
; Use:	my_xy, undo_my_xy
;
save_map:
		push	es
		push	cx
		push	ds
		pop	es
		mov	si,my_xy
		mov	di,undo_my_xy
		mov	cx,MAX_FIELD_W*MAX_FIELD_H+2
	rep	movsb
		pop	cx
		pop	es
		ret
;=============================================================================
; move
;-----------------------------------------------------------------------------
; .
; In:	---
; Out:	---
; Modf:	---
; Call:	get_val, save_map, move_box, get_addr
; Use:	my_xy, map_width, map_height, undo_enabled, moves
;
move:
		mov	dx,[my_xy]
		mov	bp,dx
		mov	cx,ax
		add	dl,al
		add	dh,ah

		or	dl,dl
		js	@@quit
		mov	al,[map_width]
		cmp	dl,al
		je	@@quit
		or	dh,dh
		js	@@quit
		mov	al,[map_height]
		cmp	dh,al
		je	@@quit

		call	get_val
		and	al,03h
		cmp	al,WALL
		je	@@quit

		mov	byte [undo_enabled],1

		cmp	al,BOX0
		je	@@mv_box
		call	save_map
		jmp	@@quit_2
@@mv_box:
		push	dx
		add	dl,cl
		add	dh,ch
		call	move_box
		pop	dx
		jc	@@quit

@@quit_2:
		call	get_addr
		and	byte [bx],W_MASK

		inc	word [moves]
		mov	[my_xy],dx
		ret
@@quit:
		mov	dx,bp
		ret

;=============================================================================
; move_box
;-----------------------------------------------------------------------------
; .
; In:	DX -- coord.
; Out:	---
; Modf:	---
; Call:	get_val, save_map, get_addr
; Use:	map_width, map_height, pushes
;
move_box:
		or	dl,dl
		js	@@quit
		mov	al,[map_width]
		cmp	dl,al
		je	@@quit
		or	dh,dh
		js	@@quit
		mov	al,[map_height]
		cmp	dh,al
		je	@@quit

		call	get_val
		and	al,03h
		cmp	al,SPACE
		jne	@@quit
		call	save_map
		call	get_addr
		and	byte [bx],W_MASK
		or	byte [bx],BOX0
		inc	word [pushes]
		clc
		ret
@@quit:
		stc
		ret
;=============================================================================
; undo
;-----------------------------------------------------------------------------
; Undoes a move.
; In:	---
; Out:	---
; Modf:	SI, DI
; Call:	---
; Use:	undo_my_xy, my_xy, undo_enabled
;
undo:
		cmp	byte [undo_enabled],0
		je	@@quit
		mov	byte [undo_enabled],0
		push	es
		push	cx
		push	ds
		pop	es
		mov	si,undo_my_xy
		mov	di,my_xy
		mov	cx,MAX_FIELD_W*MAX_FIELD_H+2
	rep	movsb
		pop	cx
		pop	es
@@quit:
		ret
;=============================================================================
; init_map
;-----------------------------------------------------------------------------
; Initializes the game map.
; In:	---
; Out:	---
; Modf:	AX, BX, CX, DI, SI, BP
; Call:	unpack
; Use:	boxes, in_place, maze_n, mazes, map_width, field_x, map_height,
;	field_y, my_xy, map
;
init_map:
		push	es
		push	ds
		pop	es
		mov	byte [boxes],0
		mov	byte [in_place],0
		mov	bx,[maze_n]
		shl	bx,1
		mov	si,[mazes+bx]
		lodsb
		cbw
		mov	[map_width],ax
		neg	ax
		add	ax,40		;Screen width / 2
		mov	[field_x],ax
		lodsb
		cbw
		mov	[map_height],ax
		push	ax
		neg	ax
		add	ax,24		;Screen height
		shr	ax,1
		mov	[field_y],ax
		lodsw
		mov	[my_xy],ax
		pop	ax
		mul	byte [map_width]
		mov	bp,ax
		mov	di,map
		add	bp,di

@@loop:
		lodsb
		mov	ch,al
		and	ch,1fh
		cmp	al,0e0h
		ja	@@lodsb
		mov	ch,1
@@loop_2:
		dec	si
@@lodsb:
		lodsb

		call	unpack
		je	@@end_loop

		call	unpack
		je	@@end_loop

		call	unpack
		je	@@end_loop

		dec	ch
		jnz	@@loop_2

		jmp	@@loop
@@end_loop:
		pop	es
		ret
;=============================================================================
; unpack
;-----------------------------------------------------------------------------
; Partially unpacks compressed byte.
; In:	AL -- byte; BP -> end of map
; Out:	ZF == 1 if end of map reached; byte [DI] contains unpacked byte
; Modf:	AX, CX, DI
; Call:	---
; Use:	---
;
unpack:
		mov	cl,6
		mov	ah,0
		div	cl
		mov	[di],ah
		inc	di
		cmp	di,bp
		ret
;=============================================================================
; print
;-----------------------------------------------------------------------------
; Prints ASCIIZ-string.
; In:	DS:SI -> string
;	AH -- color
;	ES:DI -> videobuffer
; Out:	---
; Modf:	AX (AL == 0), SI
; Call:	---
; Use:	---
;
print:
		push	di
@@next:
		lodsb
		or	al,al
		jz	@@quit
		stosw
		jmp	@@next
@@quit:
		pop	di
		ret
;=============================================================================
; fill_bg
;-----------------------------------------------------------------------------
fill_bg:
		mov	cx,1500
@@loop_00:
		push	cx
		mov	bx,[map_height]
		inc	bx		;Frame
		inc	bx
		call	random
		add	ax,[field_y]
		dec	ax
		mov	bx,80
		mul	bx
		mov	di,ax
		mov	bx,[map_width]
		inc	bx		;Frame
		inc	bx
		call	random
		dec	ax
		shl	ax,1
		add	ax,[field_x]
		add	di,ax
		shl	di,1

		call	put_rnd_char
		call	put_rnd_char

		mov	cx,700
@@l:
		in	al,80h
		loop	@@l

		pop	cx
		loop	@@loop_00
		ret
;------------
put_rnd_char:
		mov	bx,0ffffh
		call	random
		push	bx
		mov	bx,ax
		rol	bx,1
		xor	ax,bx
		mov	bx,background
		and	al,03h
		xlat
		pop	bx
		test	ah,1
		jnz	@@label_01
		mov	ah,COL_BG1
		jmp	@@label_02
@@label_01:
		mov	ah,COL_BG2
@@label_02:
		stosw
		ret
;=============================================================================
; print_frame
;-----------------------------------------------------------------------------
; Draws frame around the game field.
; In:	---
; Out:	---
; Modf:	AX, BX, CX, DX, DI, SI
; Call:	random, print_square, print
; Use:	background, map_width, map_height, help
;
print_frame:
;-----------------------------------------------------------------------------
; Fill background
;-----------------------------------------------------------------------------
		xor	di,di
		mov	cx,25*80
@@loop_01:
		call	put_rnd_char
		loop	@@loop_01
;-----------------------------------------------------------------------------
; Draw frame
;-----------------------------------------------------------------------------
		mov	dx,0ffffh	;Coordinates: DH == -1, DL == -1
		mov	al,FRAME
		mov	cx,[map_width]
		inc	cx
		inc	cx
@@loop_02:
		call	print_square
		push	dx
		mov	dh,[map_height]
		call	print_square
		pop	dx
		inc	dl
		loop	@@loop_02

		mov	dx,00ffh
		mov	cx,[map_height]
@@loop_03:
		call	print_square
		push	dx
		mov	dl,[map_width]
		call	print_square
		pop	dx
		inc	dh
		loop	@@loop_03

		mov	ax,[field_y]
		add	ax,[map_height]
		mov	ah,160
		mul	ah
		add	ax,4
		add	ax,[field_x]
		add	ax,[field_x]
		mov	di,ax

		mov	ax,[maze_n]
		inc	ax
		call	itoa
		inc	si
		inc	si
		call	print

		mov	di,24*160	;Last line
		mov	si,help
		mov	ah,COL_HELP
		call	print
		ret
;=============================================================================
; print_square
;-----------------------------------------------------------------------------
; Draws a square.
; In:	AL -- type:
;		0 -- empty
;	DX -- coordinates
; Out:	---
; Modf:	---
; Call:	---
; Use:	chars
;
print_square:
		push	ax
		push	dx
		push	di
		push	si
		add	dh,[field_y]	;Start row
		shl	dl,1
		add	dl,[field_x]	;Start column
;-----------------------------------------------------------------------------
; Compute address of char w/ coordinates in DX
;-----------------------------------------------------------------------------
		xchg	di,ax
		mov	al,dh
		mov	dh,160		;Bytes in line
		mul	dh
		xchg	di,ax
		xor	dh,dh
		shl	dx,1
		add	di,dx
;-----------------------------------------------------------------------------
; Retrieve chars and color
;-----------------------------------------------------------------------------
		mov	dx,ax
		cbw
		shl	ax,1
		add	al,dl		;AX *= 3
		add	ax,chars
		xchg	si,ax		;SI -> table w/ colors and chars
		lodsb			;Color
		xchg	al,ah
		lodsb			;1st char
		stosw
		lodsb			;2nd char
		stosw
		pop	si
		pop	di
		pop	dx
		pop	ax
		ret
;=============================================================================
; print_maze
;-----------------------------------------------------------------------------
; Shows the game field.
; In:	---
; Out:	---
; Modf:	---
; Call:	getmval, print_square
; Use:	map, map_width, map_height
;
print_maze:
		push	ax
		push	cx
		push	dx
		push	si
		mov	si,map
		mov	byte [boxes],0
		mov	byte [in_place],0

		xchg	cx,ax
		xor	dx,dx
@@next:
		lodsb
		cmp	al,BOX0
		jne	@@ne
		inc	byte [boxes]
@@ne:
		cmp	al,BOX1
		jne	@@ne2
		inc	byte [boxes]
		inc	byte [in_place]
@@ne2:
		call	print_square
		inc	dx		;OPTIMIZE: instead of INC DL
		cmp	dl,[map_width]
		jb	@@next
		xor	dl,dl
		inc	dh
		cmp	dh,[map_height]
		jb	@@next
@@quit:
		mov	dx,[my_xy]
		call	get_val
		cmp	al,TARGET
		mov	al,MAN
		jne	@@pr
		inc	ax		;MAN->MAN2
@@pr:
		call	print_square
		pop	si
		pop	dx
		pop	cx
		pop	ax
		ret
;=============================================================================
; showstatus
;-----------------------------------------------------------------------------
; .
; In:	---
; Out:	---
; Modf:	---
; Call:	itoa, print
; Use:	field_y, field_x, moves, pushes, starttime
;
showstatus:
		push	ax
		push	dx
		push	di
		push	si

		mov	ax,[field_y]
		mov	ah,160
		dec	al
		mul	ah
		add	ax,4
		add	ax,[field_x]
		add	ax,[field_x]
		mov	di,ax

		mov	ax,[moves]
		call	itoa
		call	print

		mov	ax,[pushes]
		call	itoa
		add	di,12
		call	print		;After print AL == 0

		cbw			;OPTIMIZE: instead of MOV AH,0
		int	1ah
		sub	dx,[starttime]
		sbb	cx,[starttime+2]
		mov	al,0		;Shift time left by 8 bits
		mov	ah,dl
		mov	dl,dh
		mov	dh,cl
		mov	si,4659		;Timer ticks per 256 sec.
		div	si

		call	itoa
		add	di,12
		call	print
		pop	si
		pop	di
		pop	dx
		pop	ax
		ret
;=============================================================================
; changefield
;-----------------------------------------------------------------------------
; Increaces/decreaces size of the field.
; In:	AX -- keyboard codes from int 16h, fn 00h
; Out:	---
; Modf:	AX, DI, SI
; Call:	---
; Use:	maze_n
;
changefield:
		mov	di,[maze_n]
		clc
		cmp	al,'-'
		je	@@decrement
		cmp	di,LAST_MAZE
		je	@@stc
		inc	di
		mov	[maze_n],di
		clc
		ret
@@decrement:
		cmp	di,0
		je	@@stc
		dec	di
		mov	[maze_n],di
		clc
		ret
@@stc:
		stc
		ret
;=============================================================================
; get_addr
;-----------------------------------------------------------------------------
; Returns address of map element.
; In:	DX -- coordinates
; Out:	BX -> element
; Modf:	!BX
; Call:	---
; Use:	map, map_width
;
get_addr:
		push	ax
		mov	bx,map
		mov	ax,[map_width]
		mul	dh
		add	al,dl
		adc	ah,0
		add	bx,ax
		pop	ax
		ret
;=============================================================================
; get_val
;-----------------------------------------------------------------------------
; Returns value of element in the map.
; In:	DX -- coordinates
; Out:	AL -- value
; Modf:	!AX
; Call:	get_addr
; Use:	---
;
get_val:
		push	bx
		call	get_addr
		mov	al,[bx]
		pop	bx
		ret
;=============================================================================
; random
;-----------------------------------------------------------------------------
; Generates a random number.
; In:	BX -- limit
; Out:	AX -- random number
; Modf:	!AX
; Call:	---
; Use:	randnum
;
random:
		push	bx
		push	cx
		push	dx
		mov	cx,[randnum]
		mov	ax,[randnum+2]
		mov	ah,al
		mov	al,ch
		mov	ch,cl
		rcr	ax,1
		rcr	cx,1
		add	[randnum],cx
		adc	ax,[randnum+2]
		add	word [randnum],62e9h
		adc	ax,3619h
		mov	[randnum+2],ax
		xor	dx,dx
		or	bx,bx
		jz	@@skip
		div	bx
@@skip:
		xchg	ax,dx
		pop	dx
		pop	cx
		pop	bx
		ret
;=============================================================================
; itoa
;-----------------------------------------------------------------------------
; Converts integer number w/ sign to ASCIIZ-string up to 5 bytes (with \0).
; In:	AX -- number
; Out:	AH == COL_STATUS
;		AL == 0
;		SI -> string
;		`cnvtbuf' contains character representaion of number
; Modf:	!AX, !SI
; Call:	---
; Use:	cnvtbuf
;
itoa:
		push	cx
		push	dx
		mov	si,cnvtbuf
		push	si
		mov	cx,4
		or	ax,ax
		jns	@@loop_01
		mov	byte [si],'-'
		neg	ax
		dec	cx
		inc	si
@@loop_01:
		mov	byte [si],'0'
		inc	si
		loop	@@loop_01
		mov	byte [si],0
		or	ax,ax
		jz	@@quit
		mov	cx,10
@@loop_02:
		xor	dx,dx
		div	cx
		add	dl,'0'
		dec	si
		mov	[si],dl
		or	ax,ax
		jnz	@@loop_02
@@quit:
		mov	ah,COL_STATUS
		pop	si
		pop	dx
		pop	cx
		ret
;=============================================================================
; charxchg
;-----------------------------------------------------------------------------
; Exchanges chars from character generator and from data segment.
; Idea from SNAKE.COM by PhysTechSoft.
; In:	---
; Out:	---
; Modf:	AX, CX, DX, DI, SI
; Call:	outport
; Use:	open, font, close
;
charxchg:
		push	ds
		cld
		mov	dx,3dah
@@wait_retrace:
		in	al,dx
		test	al,8
		jz	@@wait_retrace
		mov	si,open		;To open character generator
		call	outport
		mov	si,0c0h * 32	;Offset of char 0c0h
		mov	cx,18		;18 chars to exchange
		mov	ax,0a000h
		mov	ds,ax
		mov	di,font
@@char_loop:
		push	cx
		mov	cx,16		;16 bytes per char
@@byte_loop:
		mov	ah,[cs:di]	;rep xchg byte [cs:di],byte [ds:si]
		xchg	ah,[si]		;-)
		mov	[cs:di],ah
		cmpsb			;OPTIMIZE: instead of INC SI; INC DI
		loop	@@byte_loop

		add	si,16		;SI -> next char
		pop	cx
		loop	@@char_loop

		pop	ds
		mov	si,close
		call	outport
		ret
;=============================================================================
; outport
;-----------------------------------------------------------------------------
; Outputs values needed to open and close the character generator to VGA
; regs 3c4h and 3ceh
; In:	SI -> table: 2 words for 3c4h, 3 words for 3ceh
; Out:	---
; Modf:	AX, DX, SI
; Call:	---
; Use:	---
;
outport:
		mov	dx,3c4h
		cli
		lodsw
		out	dx,ax
		lodsw
		out	dx,ax
		sti
		mov	dl,0ceh
		lodsw
		out	dx,ax
		lodsw
		out	dx,ax
		lodsw
		out	dx,ax
		ret
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
		section	.data align=1
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Initialized data
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Text strings
;
copyright	db	"Pusher ", RELEASE, CR, LF
		db	"Copyright (c) 2009 Oleg O. Chukaev <oleg.chukaev@mail.ru>", CR, LF
		db	"This program is free software; you can redistribute it and/or modify", CR, LF
		db	"it under the terms of the GNU General Public License as published by", CR, LF
		db	"the Free Software Foundation; either version 2 of the License, or", CR, LF
		db	"(at your option) any later version.", CR, LF, "$"
help		db	"Keys: ", 18h, " ", 19h, " ", 1bh, " ", 1ah, " [R]estart/Space [U]ndo/BkSp"
		db	" + - [Q]uit/ESC     Pusher  2009 0leg", 0
background	db	"-"		;Chars to fill backgroung
;
;-----------------------------------------------------------------------------
; Misc. variables
;
maze_n		dw	0
;
;-----------------------------------------------------------------------------
; Squares' codes and chars
;
chars		db	COL_EMPTY, 0c0h, 0c1h	;space
		db	COL_BOX0, 0c6h, 0c7h	;box 0
		db	COL_WALL, 0c8h, 0c9h	;wall
		db	COL_EMPTY, 0c0h, 0c1h	;empty
		db	COL_TARGET, 0d0h, 0d1h	;target
		db	COL_BOX1, 0c6h, 0c7h	;box 1
		db	COL_FRAME, 0cah, 0cbh	;frame
		db	COL_MAN, 0c2h, 0c3h	;man
		db	COL_MAN, 0c4h, 0c5h	;man 2
;
;-----------------------------------------------------------------------------
; Additions to cursor coordinates
;
directions	db	KEY_LEFT, -1,  0, KEY_RIGHT, 1, 0
		db	KEY_UP,    0, -1, KEY_DOWN,  0, 1
		db	0		;End of table
;
;-----------------------------------------------------------------------------
; Data for opening/closing the character generator
;
open		dw	0402h, 0704h, 0204h, 0005h, 0006h
close		dw	0302h, 0304h, 0004h, 1005h, 0e06h
;
;-----------------------------------------------------------------------------
; Mazes
;
%include	"maze.inc"
;
;-----------------------------------------------------------------------------
; Font
;
font:
;-----------------------------------------------------------------------------
; 0c0h Empty cell -- left side
;-----------------------------------------------------------------------------
		db	01010101b	;1
		db	10000000b	;2
		db	00000000b	;3
		db	10000000b	;4
		db	00000000b	;5
		db	10000000b	;6
		db	00000000b	;7
		db	10000000b	;8
		db	00000000b	;9
		db	10000000b	;10
		db	00000000b	;11
		db	10000000b	;12
		db	00000000b	;13
		db	10000000b	;14
		db	01010101b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0c1h Empty cell -- right side
;-----------------------------------------------------------------------------
		db	01010100b	;1
		db	00000010b	;2
		db	00000000b	;3
		db	00000010b	;4
		db	00000000b	;5
		db	00000010b	;6
		db	00000000b	;7
		db	00000010b	;8
		db	00000000b	;9
		db	00000010b	;10
		db	00000000b	;11
		db	00000010b	;12
		db	00000000b	;13
		db	00000010b	;14
		db	01010100b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0c2h The Man -- left side
;-----------------------------------------------------------------------------
		db	~11111111b	;1
		db	~11000000b	;2
		db	~10000000b	;3
		db	~10011000b	;4
		db	~10111100b	;5
		db	~10011000b	;6
		db	~10000000b	;7
		db	~10000000b	;8
		db	~10000000b	;9
		db	~10000000b	;10
		db	~10110000b	;11
		db	~10111111b	;12
		db	~10001111b	;13
		db	~10000000b	;14
		db	~11000000b	;15
		db	~11111111b	;16
;-----------------------------------------------------------------------------
; 0c3h The Man -- right side
;-----------------------------------------------------------------------------
		db	~11111111b	;1
		db	~00000011b	;2
		db	~00000001b	;3
		db	~00011001b	;4
		db	~00111101b	;5
		db	~00011001b	;6
		db	~00000001b	;7
		db	~00000001b	;8
		db	~00000001b	;9
		db	~00000001b	;10
		db	~00001101b	;11
		db	~11111101b	;12
		db	~11110001b	;13
		db	~00000001b	;14
		db	~00000011b	;15
		db	~11111111b	;16
;-----------------------------------------------------------------------------
; 0c4h The Man 2 -- left side
;-----------------------------------------------------------------------------
		db	~11111111b	;1
		db	~11000000b	;2
		db	~10000000b	;3
		db	~10011100b	;4
		db	~10100010b	;5
		db	~10000000b	;6
		db	~10000000b	;7
		db	~10000000b	;8
		db	~10000000b	;9
		db	~10000000b	;10
		db	~10110000b	;11
		db	~10111111b	;12
		db	~10001111b	;13
		db	~10000000b	;14
		db	~11000000b	;15
		db	~11111111b	;16
;-----------------------------------------------------------------------------
; 0c5h The Man 2 -- right side
;-----------------------------------------------------------------------------
		db	~11111111b	;1
		db	~00000011b	;2
		db	~00000001b	;3
		db	~00111001b	;4
		db	~01000101b	;5
		db	~00000001b	;6
		db	~00000001b	;7
		db	~00000001b	;8
		db	~00000001b	;9
		db	~00000001b	;10
		db	~00001101b	;11
		db	~11111101b	;12
		db	~11110001b	;13
		db	~00000001b	;14
		db	~00000011b	;15
		db	~11111111b	;16
;-----------------------------------------------------------------------------
; 0c6h Box -- left side
;-----------------------------------------------------------------------------
		db	11111111b	;1
		db	10000000b	;2
		db	11000100b	;3
		db	11000100b	;4
		db	10000000b	;5
		db	10100010b	;6
		db	10100010b	;7
		db	10000000b	;8
		db	11000100b	;9
		db	11000100b	;10
		db	10000000b	;11
		db	10100010b	;12
		db	10100010b	;13
		db	10000000b	;14
		db	11111111b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0c7h Box -- right side
;-----------------------------------------------------------------------------
		db	11111110b	;1
		db	00000010b	;2
		db	10001010b	;3
		db	10001010b	;4
		db	00000010b	;5
		db	01000110b	;6
		db	01000110b	;7
		db	00000010b	;8
		db	10001010b	;9
		db	10001010b	;10
		db	00000010b	;11
		db	01000110b	;12
		db	01000110b	;13
		db	00000010b	;14
		db	11111110b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0c8h Wall -- left side
;-----------------------------------------------------------------------------
		db	11111111b	;1
		db	10010001b	;2
		db	10001000b	;3
		db	11010000b	;4
		db	11000010b	;5
		db	10001010b	;6
		db	11000001b	;7
		db	10100000b	;8
		db	10001101b	;9
		db	11100000b	;10
		db	11000000b	;11
		db	10100110b	;12
		db	10111111b	;13
		db	11111111b	;14
		db	11111111b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0c9h Wall -- right side
;-----------------------------------------------------------------------------
		db	11111110b	;1
		db	10000110b	;2
		db	00101110b	;3
		db	00100110b	;4
		db	10011110b	;5
		db	00101110b	;6
		db	00000110b	;7
		db	11010110b	;8
		db	00000110b	;9
		db	00100110b	;10
		db	10011110b	;11
		db	01001110b	;12
		db	11111110b	;13
		db	11111110b	;14
		db	11111110b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0cah Border -- left side
;-----------------------------------------------------------------------------
		db	00101000b	;1
		db	00000001b	;2
		db	00001000b	;3
		db	01010000b	;4
		db	01000010b	;5
		db	00001010b	;6
		db	01000001b	;7
		db	00100000b	;8
		db	00001101b	;9
		db	01100000b	;10
		db	01000000b	;11
		db	00100110b	;12
		db	00111111b	;13
		db	01111111b	;14
		db	01111111b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0cbh Border -- right side
;-----------------------------------------------------------------------------
		db	01010110b	;1
		db	10000110b	;2
		db	00101110b	;3
		db	00100110b	;4
		db	10011110b	;5
		db	00101110b	;6
		db	00000110b	;7
		db	11010110b	;8
		db	00000110b	;9
		db	00100110b	;10
		db	10011110b	;11
		db	01001110b	;12
		db	11111110b	;13
		db	11111110b	;14
		db	11111110b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0cch (c) -- left side
;-----------------------------------------------------------------------------
		db	00000000b	;1
		db	00000000b	;2
		db	00000111b	;3
		db	00011100b	;4
		db	00110000b	;5
		db	00110011b	;6
		db	01100110b	;7
		db	01100110b	;8
		db	01100110b	;9
		db	01100110b	;10
		db	00110011b	;11
		db	00110000b	;12
		db	00011100b	;13
		db	00000111b	;14
		db	00000000b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0cdh (c) -- right side
;-----------------------------------------------------------------------------
		db      00000000b	;1
		db      00000000b	;2
		db	11100000b	;3
		db	00111000b	;4
		db	00001100b	;5
		db	11001100b	;6
		db	01100110b	;7
		db	00000110b	;8
		db	00000110b	;9
		db	01100110b	;10
		db	11001100b	;11
		db	00001100b	;12
		db	00111000b	;13
		db	11100000b	;14
		db      00000000b	;15
		db      00000000b	;16
;-----------------------------------------------------------------------------
; 0ceh Screen background
;-----------------------------------------------------------------------------
		db	00000000b	;1
		db	00000000b	;2
		db	00000000b	;3
		db	00000000b	;4
		db	00001000b	;5
		db	00001000b	;6
		db	00001000b	;7
		db	00001000b	;8
		db	00001000b	;9
		db	00001000b	;10
		db	00001000b	;11
		db	00001000b	;12
		db	00000000b	;13
		db	00000000b	;14
		db	00000000b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0cfh Screen background
;-----------------------------------------------------------------------------
		db	00000000b	;1
		db	00001000b	;2
		db	00001000b	;3
		db	00001000b	;4
		db	00001000b	;5
		db	00001000b	;6
		db	00001000b	;7
		db	00001000b	;8
		db	00001000b	;9
		db	00001000b	;10
		db	00001000b	;11
		db	00001000b	;12
		db	00001000b	;13
		db	00001000b	;14
		db	00001000b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0d0h Target -- left side
;-----------------------------------------------------------------------------
		db	00000010b	;1
		db	00000000b	;2
		db	00001000b	;3
		db	00000000b	;4
		db	00100000b	;5
		db	00000000b	;6
		db	10000000b	;7
		db	00000000b	;8
		db	10000000b	;9
		db	00000000b	;10
		db	00100000b	;11
		db	00000000b	;12
		db	00001000b	;13
		db	00000000b	;14
		db	00000010b	;15
		db	00000000b	;16
;-----------------------------------------------------------------------------
; 0d1h Target -- right side
;-----------------------------------------------------------------------------
		db	10000000b	;1
		db	00000000b	;2
		db	00100000b	;3
		db	00000000b	;4
		db	00001000b	;5
		db	00000000b	;6
		db	00000010b	;7
		db	00000000b	;8
		db	00000010b	;9
		db	00000000b	;10
		db	00001000b	;11
		db	00000000b	;12
		db	00100000b	;13
		db	00000000b	;14
		db	10000000b	;15
		db	00000000b	;16
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
		section	.bss align=1
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; Uninitialized data
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;
save_buffer	resw	25*80		;Buffer for screen contents
moves		resw	1		;
pushes		resw	1		;
boxes		resb	1		;
in_place	resb	1		;
field_x		resw	1		;
field_y		resw	1		;
map_width	resw	1		;Field's width
map_height	resw	1		;Field's height
undo_enabled	resb	1		;
starttime	resw	2		;Time when game started
curshape	resw	1		;Saved sursor shape
randnum		resw	2		;Random number
rgb		resb	3		;Saved palette
cnvtbuf		resb	5		;Buffer for `itoa'
my_xy		resw	1		;
map		resb	MAX_FIELD_W*MAX_FIELD_H
undo_my_xy	resw	1		;
undo_map	resb	MAX_FIELD_W*MAX_FIELD_H
;
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
; E0F
;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


